import os
import os.path
import shutil

import click
import magic
from doit.doit_cmd import DoitMain
from doit.cmd_base import ModuleTaskLoader

from elm_doc.tasks import create_tasks


class DoitException(click.ClickException):
    def __init__(self, message, exit_code):
        click.ClickException.__init__(self, message)
        self.exit_code = exit_code


def validate_mount_at(ctx, param, value):
    if len(value) == 0 or value.startswith('/'):
        return value
    else:
        raise click.BadParameter('should be an absolute path, like /{}'.format(value))


def validate_elm_make(ctx, param, value):
    if value is None:
        return value

    realpath = os.path.realpath(value)
    if not os.path.isfile(realpath):
        realpath = shutil.which(value)

    if realpath is None or not os.path.isfile(realpath):
        raise click.BadParameter('{} not found'.format(value))

    elm_make_mimetype = magic.from_file(realpath, mime=True)
    if not elm_make_mimetype.startswith('text'):
        return value

    perhaps_binwrap_of = os.path.normpath(
        os.path.join(
            os.path.dirname(realpath),
            os.pardir,
            'elm',
            'Elm-Platform',
            '*',
            '.cabal-sandbox',
            'bin',
            'elm-make'))
    raise click.BadParameter('''should be the real elm-make binary; this looks like a text file.
if you installed Elm through npm, then try {}'''.format(perhaps_binwrap_of))


@click.command(context_settings=dict(
    ignore_unknown_options=True,
))
@click.option('--output', '-o',
              metavar='dir')
@click.option('--elm-make',
              metavar='path/to/elm-make',
              callback=validate_elm_make,
              help=('specify which elm-make to use. if not specified, '
                    'elm will be installed afresh in a temporary directory'))
@click.option('--mount-at',
              metavar='/path',
              default='',
              callback=validate_mount_at,
              help='url path at which the docs will be served')
@click.option('--exclude', '-x',
              metavar='module1,module2.*',
              help='comma-separated fnmatch pattern of modules to exclude')
@click.option('--validate/--no-validate',
              default=False,
              help='validate all doc comments are in place without generating docs')
@click.argument('project_path')
@click.argument('doit_args', nargs=-1, type=click.UNPROCESSED)
def main(output, elm_make, mount_at, exclude, validate, project_path, doit_args):
    """Generate static documentation for your Elm project"""

    if not validate and output is None:
        raise click.BadParameter('please specify --output directory')

    def task_build():
        exclude_modules = exclude.split(',') if exclude else []
        return create_tasks(
            os.path.abspath(project_path),
            os.path.abspath(output) if output is not None else None,
            elm_make=os.path.abspath(elm_make) if elm_make is not None else None,
            exclude_modules=exclude_modules,
            mount_point=mount_at,
            validate=validate)

    result = DoitMain(ModuleTaskLoader(locals())).run(doit_args)
    if result is not None and result > 0:
        raise DoitException('', result)


if __name__ == '__main__':
    main()



'''
'''
from typing import List, Optional
import os.path
from pathlib import Path

from elm_doc import elm_package
from elm_doc import package_tasks
from elm_doc import asset_tasks
from elm_doc import catalog_tasks


def create_tasks(
        project_path: str,
        output_dir: Optional[str] = None,
        elm_make: Optional[str] = None,
        exclude_modules: List[str] = [],
        mount_point: str = '',
        validate: bool = False):
    output_path = Path(os.path.normpath(output_dir)) if output_dir is not None else None
    elm_make = Path(os.path.normpath(elm_make)) if elm_make is not None else None
    # todo: gracefully handle missing elm-package.json
    project_package = elm_package.from_path(Path(os.path.abspath(os.path.normpath(project_path))))
    # todo: gracefully handle missing exact-dependencies.json
    deps = list(elm_package.iter_dependencies(project_package))
    all_packages = [project_package] + deps

    for task in package_tasks.create_package_tasks(
            output_path,
            project_package,
            elm_make=elm_make,
            exclude_modules=exclude_modules,
            mount_point=mount_point,
            validate=validate):
        yield task

    if validate:
        return

    for package in deps:
        for task in package_tasks.create_package_tasks(
                output_path, package, elm_make=elm_make, mount_point=mount_point):
            yield task

    for task in catalog_tasks.create_catalog_tasks(all_packages, output_path, mount_point=mount_point):
        yield task

    yield {
        'basename': 'assets',
        'actions': [(asset_tasks.build_assets, (output_path, mount_point))],
        'targets': [output_path / 'assets', output_path / 'artifacts'],
        'uptodate': [True],
    }



from typing import List
from pathlib import Path
import json

from elm_doc.elm_package import ElmPackage
from elm_doc import elm_package
from elm_doc import page_template


def write_index_page(output_path: Path, mount_point: str = ''):
    with open(str(output_path), 'w') as f:
        f.write(page_template.render('Catalog', mount_point=mount_point))


def write_all_packages(packages: List[ElmPackage], output_path: Path):
    all_packages = map(
        lambda package: {
            'name': package.name,
            'summary': package.summary,
            'versions': [package.version],
        },
        packages)
    with open(str(output_path), 'w') as f:
        json.dump(list(all_packages), f)


def write_new_packages(packages: List[ElmPackage], output_path: Path):
    new_packages = map(lambda package: package.name, packages)
    with open(str(output_path), 'w') as f:
        json.dump(list(new_packages), f)


def create_catalog_tasks(packages: List[ElmPackage], output_path: Path, mount_point: str = ''):
    # index
    index_path = output_path / 'index.html'
    yield {
        'basename': 'index',
        'actions': [(write_index_page, (index_path, mount_point))],
        'targets': [index_path],
        'uptodate': [True],
    }

    # all-packages
    all_packages_path = output_path / 'all-packages'
    yield {
        'basename': 'all_packages',
        'actions': [(write_all_packages, (packages, all_packages_path))],
        'targets': [all_packages_path],
        'file_dep': list(map(elm_package.description_path, packages)),
    }

    # new-packages
    new_packages_path = output_path / 'new-packages'
    yield {
        'basename': 'new_packages',
        'actions': [(write_new_packages, (packages, new_packages_path))],
        'targets': [new_packages_path],
        'file_dep': list(map(elm_package.description_path, packages)),
    }



import imp
from distutils.sysconfig import get_python_lib


def elm_package_overlayer_path():
    search_paths = __path__ + [get_python_lib()]
    (_file, overlayer_path, _desc) = imp.find_module(
        "overlay_elm_package", search_paths)
    return overlayer_path



from typing import NamedTuple, Dict, List, Iterator
from pathlib import Path
import fnmatch
import json
import urllib.parse


ModuleName = str
DESCRIPTION_FILENAME = 'elm-package.json'
STUFF_DIRECTORY = 'elm-stuff'
EXACT_DEPS_FILENAME = 'exact-dependencies.json'
PACKAGES_DIRECTORY = 'packages'


ElmPackageBase = NamedTuple('ElmPackageBase', [
    ('path', Path),
    ('description', Dict),
    ('name', str),
    ('user', str),
    ('project', str),
    ('is_dep', bool),
])


class ElmPackage(ElmPackageBase):
    def __getattr__(self, name):
        return self.description[name.replace('_', '-')]


def from_path(path: Path, is_dep: bool = False) -> ElmPackage:
    description = load_description(path / DESCRIPTION_FILENAME)
    repo_path = Path(urllib.parse.urlparse(description['repository']).path)
    user = repo_path.parent.stem
    project = repo_path.stem
    return ElmPackage(
        path=path,
        description=description,
        name='{0}/{1}'.format(user, project),
        user=user,
        project=project,
        is_dep=is_dep,
    )


def description_path(package: ElmPackage) -> Path:
    return package.path / DESCRIPTION_FILENAME


def load_description(path: Path) -> Dict:
    with open(str(path)) as f:
        return json.load(f)


def iter_dependencies(package: ElmPackage) -> Iterator[ElmPackage]:
    exact_deps_path = package.path / STUFF_DIRECTORY / EXACT_DEPS_FILENAME
    exact_deps = {}
    try:
        with open(str(exact_deps_path)) as f:
            exact_deps = json.load(f)
    except OSError:
        # todo: warn about missing exact deps
        pass
    for name, version in exact_deps.items():
        yield from_path(package.path / STUFF_DIRECTORY / PACKAGES_DIRECTORY / name / version, is_dep=True)


def glob_package_modules(package: ElmPackage, exclude_patterns: List[str] = []) -> Iterator[ModuleName]:
    for source_dir_name in package.source_directories:
        source_dir = package.path / source_dir_name
        elm_files = source_dir.glob('**/*.elm')
        for elm_file in elm_files:
            if elm_file.relative_to(package.path).parts[0] == STUFF_DIRECTORY:
                continue
            rel_path = elm_file.relative_to(source_dir)
            module_name = '.'.join(rel_path.parent.parts + (rel_path.stem,))
            if any(fnmatch.fnmatch(module_name, pattern)
                   for pattern in exclude_patterns):
                continue
            yield module_name



import json


PAGE_TEMPLATE = '''
<!DOCTYPE html>
<html>
  <head>
    <meta charset="UTF-8">
    <link rel="shortcut icon" size="16x16, 32x32, 48x48, 64x64, 128x128, 256x256" href="{mount_point}/assets/favicon.ico">
    <link rel="stylesheet" href="{mount_point}/assets/highlight/styles/default.css">
    <link rel="stylesheet" href="{mount_point}/assets/style.css">
    <script src="{mount_point}/assets/highlight/highlight.pack.js"></script>
    <script src="{mount_point}/artifacts/Page-{page_module}.js"></script>
  </head>
  <body>
  <script>
    var page = Elm.Page.{page_module}.fullscreen({flags_json});
  </script>
  </body>
</html>
'''  # noqa: E501


def render(page_module: str, flags: dict = None, mount_point: str = ''):
    flags_json = json.dumps(flags) if flags else ''
    return PAGE_TEMPLATE.format(
        mount_point=mount_point, page_module=page_module, flags_json=flags_json)



import os.path
from tempfile import TemporaryDirectory
import tarfile
import subprocess
from pathlib import Path

import conftest
from elm_doc import elm_platform, elm_package


def task_create_elm_stuff_fixture():
    elm_stuff_fixture_path = conftest.elm_stuff_fixture_path()
    elm_versions = ['0.18.0']
    for elm_version in elm_versions:
        tarball = str(elm_stuff_fixture_path(elm_version))
        yield {
            'basename': 'create_elm_stuff_fixutre:' + elm_version,
            'actions': [(create_elm_stuff_fixture, (elm_version, tarball))],
            'targets': [tarball],
            'uptodate': [True],
        }


def create_elm_stuff_fixture(elm_version: str, tarball: str):
    os.makedirs(os.path.dirname(tarball), exist_ok=True)
    with TemporaryDirectory() as tmpdir:
        root_path = Path(tmpdir)
        elm_platform.install(root_path, elm_version)
        subprocess.check_call(['./node_modules/.bin/elm-package', 'install', '--yes'], cwd=root_path)
        elm_stuff = root_path / elm_package.STUFF_DIRECTORY
        with tarfile.open(tarball, "w:gz") as tar:
            tar.add(elm_stuff, arcname=os.path.basename(elm_stuff))



import functools
import subprocess

from doit.exceptions import TaskFailed


def capture_subprocess_error(fn):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except subprocess.CalledProcessError as e:
            return TaskFailed('Error while executing {}'.format(' '.join(e.cmd)))
    return wrapper



from typing import List, Optional
import os
import os.path
from collections import ChainMap
from pathlib import Path
import json
import shutil
from tempfile import TemporaryDirectory
import subprocess
import urllib.request

from doit.tools import create_folder

from elm_doc import elm_platform
from elm_doc import elm_package_overlayer_path
from elm_doc import elm_package
from elm_doc.elm_package import ElmPackage, ModuleName
from elm_doc import page_template
from elm_doc.decorators import capture_subprocess_error


def get_page_package_flags(package: ElmPackage, module: Optional[str] = None):
    flags = {
        'user': package.user,
        'project': package.project,
        'version': package.version,
        'allVersions': [package.version],
        'moduleName': module,
    }
    return flags


def build_package_page(package: ElmPackage, output_path: Path, module: Optional[str] = None, mount_point: str = ''):
    os.makedirs(os.path.dirname(str(output_path)), exist_ok=True)
    with open(str(output_path), 'w') as f:
        f.write(page_template.render(
            'Package', flags=get_page_package_flags(package, module), mount_point=mount_point
        ))


def link_latest_package_dir(package_dir: Path, link_path: Path):
    os.makedirs(str(package_dir), exist_ok=True)
    link_path.symlink_to(package_dir, target_is_directory=True)


def copy_package_readme(package_readme: Path, output_path: Path):
    if package_readme.is_file():
        shutil.copy(str(package_readme), str(output_path))


@capture_subprocess_error
def build_package_docs_json(
        package: ElmPackage,
        package_modules: List[ModuleName],
        output_path: Path = None,
        elm_make: Path = None,
        validate: bool = False):
    elm_package_with_exposed_modules = dict(ChainMap(
        {'exposed-modules': package_modules},
        package.description,
    ))
    overlayer_path = elm_package_overlayer_path()
    with TemporaryDirectory() as tmpdir:
        root_path = Path(tmpdir)

        overlayed_elm_package_path = root_path / elm_package.DESCRIPTION_FILENAME
        with open(str(overlayed_elm_package_path), 'w') as f:
            json.dump(elm_package_with_exposed_modules, f)

        if elm_make is None:
            elm_platform.install(root_path, package.elm_version)
            elm_make = elm_platform.get_npm_executable_path(root_path, 'elm-make')

        if validate:
            # todo: support windows if we want to
            output_path = '/dev/null'

        # todo: make overlayer support windows if we want to (can we?)
        env = dict(ChainMap(
            {
                'USE_ELM_PACKAGE': str(overlayed_elm_package_path),
                'INSTEAD_OF_ELM_PACKAGE': str(elm_package.description_path(package)),
                'DYLD_INSERT_LIBRARIES': overlayer_path,
                'LD_PRELOAD': overlayer_path,
            },
            os.environ,
        ))
        subprocess.check_call(
            [str(elm_make), '--yes', '--docs', str(output_path), '--output', '/dev/null'],
            cwd=str(package.path),
            env=env)


def download_package_docs_json(package: ElmPackage, output_path: Path):
    url = 'http://package.elm-lang.org/packages/{name}/{version}/documentation.json'.format(
        name=package.name, version=package.version
    )
    urllib.request.urlretrieve(url, str(output_path))


def package_task_basename_factory(package):
    return lambda name: '{}:{}/{}'.format(name, package.name, package.version)


def create_package_tasks(
        output_path: Optional[Path],
        package: ElmPackage,
        elm_make: Path = None,
        exclude_modules: List[str] = [],
        mount_point: str = '',
        validate: bool = False):
    basename = package_task_basename_factory(package)

    if package.is_dep:
        package_modules = package.exposed_modules
    else:
        package_modules = list(elm_package.glob_package_modules(package, exclude_modules))

    if validate:
        yield {
            'basename': basename('validate_package_docs_json'),
            'actions': [(build_package_docs_json,
                         (package, package_modules),
                         {'elm_make': elm_make, 'validate': True})],
            'targets': [],
            # 'file_dep': [all_elm_files_in_source_dirs] # todo
        }
        return

    package_docs_root = output_path / 'packages' / package.user / package.project / package.version

    # package documentation.json
    docs_json_path = package_docs_root / 'documentation.json'
    if package.is_dep:
        yield {
            'basename': basename('download_package_docs_json'),
            'actions': [(create_folder, (str(package_docs_root),)),
                        (download_package_docs_json, (package, docs_json_path))],
            'targets': [docs_json_path],
            # 'file_dep': [all_elm_files_in_source_dirs] # todo
            'uptodate': [True],
        }
    else:
        yield {
            'basename': basename('build_package_docs_json'),
            'actions': [(create_folder, (str(package_docs_root),)),
                        (build_package_docs_json,
                         (package, package_modules),
                         {'elm_make': elm_make, 'output_path': docs_json_path})],
            'targets': [docs_json_path],
            # 'file_dep': [all_elm_files_in_source_dirs] # todo
        }

    # package index page
    package_index_output = package_docs_root / 'index.html'
    yield {
        'basename': basename('package_page'),
        'actions': [(build_package_page, (package, package_index_output), {'mount_point': mount_point})],
        'targets': [package_index_output],
        # 'file_dep': [module['source_file']] #todo
        'uptodate': [True],
    }

    # package readme
    readme_filename = 'README.md'
    package_readme = package.path / readme_filename
    output_readme_path = package_docs_root / readme_filename
    if package_readme.is_file():
        yield {
            'basename': basename('package_readme'),
            'actions': [(copy_package_readme, (package_readme, output_readme_path))],
            'targets': [output_readme_path],
            'file_dep': [package_readme],
        }

    # link from /latest
    latest_path = package_docs_root.parent / 'latest'
    yield {
        'basename': basename('package_latest_link'),
        'actions': [(link_latest_package_dir, (package_docs_root, latest_path))],
        'targets': [latest_path],
        # 'file_dep': [], # todo
        'uptodate': [True]
    }

    # module pages
    for module in package_modules:
        module_output = package_docs_root / module.replace('.', '-')
        yield {
            'basename': basename('module_page') + ':' + module,
            'actions': [(build_package_page, (package, module_output, module), {'mount_point': mount_point})],
            'targets': [module_output],
            # 'file_dep': [module['source_file']] #todo
            'uptodate': [True],
        }



import os
import os.path
from collections import ChainMap
from tempfile import TemporaryDirectory
from pathlib import Path
import subprocess
import shutil

from elm_doc import elm_platform
from elm_doc import elm_package
from elm_doc import node_modules
from elm_doc.decorators import capture_subprocess_error


codeshifter = os.path.normpath(os.path.join(os.path.dirname(__file__), 'native', 'prepend_mountpoint.js'))


@capture_subprocess_error
def build_assets(output_path: Path, mount_point: str = ''):
    tarball = 'https://api.github.com/repos/elm-lang/package.elm-lang.org/tarball'
    with TemporaryDirectory() as tmpdir:
        root_path = Path(tmpdir)

        # download package.elm-lang.org repo
        subprocess.check_call(
            'curl -L -s {tarball} | tar xz --strip-components 1'.format(tarball=tarball),
            shell=True,
            cwd=str(root_path),
        )
        package = elm_package.from_path(root_path)

        # install elm
        elm_platform.install(root_path, package.elm_version)
        node_modules.add('jscodeshift', cwd=str(root_path))
        subprocess.check_call(
            ['./node_modules/.bin/elm-package', 'install', '--yes'],
            cwd=str(root_path),
        )

        # make artifacts
        artifacts_path = root_path / 'artifacts'
        os.makedirs(str(artifacts_path))

        # I know there's only one source directory!
        frontend_pages = Path(package.source_directories[0]) / 'Page'
        for main_elm in root_path.glob(str(frontend_pages / '*.elm')):
            basename = main_elm.stem
            subprocess.check_call(
                ['./node_modules/.bin/elm-make',
                 str(main_elm),
                 '--output',
                 'artifacts/Page-{0}.js'.format(basename)],
                cwd=str(root_path),
            )

        env = dict(ChainMap(
            os.environ,
            {'ELM_DOC_MOUNT_POINT': mount_point},
        ))
        codeshift_command = [
            './node_modules/.bin/jscodeshift',
            '--transform',
            codeshifter,
            str(artifacts_path)]
        proc = subprocess.Popen(
            codeshift_command,
            env=env,
            cwd=str(root_path),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True)
        out, err = proc.communicate()
        # todo: jscodeshift doesn't exit with 1 when there's an error
        if ('ERROR' in out) or ('ERROR' in err) or (proc.returncode != 0):
            raise subprocess.CalledProcessError(
                proc.returncode, codeshift_command, out, err)

        # copy artifacts
        shutil.rmtree(str(output_path / 'artifacts'), ignore_errors=True)
        shutil.copytree(str(artifacts_path), str(output_path / 'artifacts'))

        # copy assets
        shutil.rmtree(str(output_path / 'assets'), ignore_errors=True)
        shutil.copytree(str(root_path / 'assets'), str(output_path / 'assets'))



"""
Generate static documentation for your Elm project.
"""
import sys
from setuptools import find_packages, setup
from setuptools.extension import Extension

if sys.platform == 'darwin':
    from distutils import sysconfig
    vars = sysconfig.get_config_vars()
    vars['LDSHARED'] = vars['LDSHARED'].replace('-bundle', '-dynamiclib')

ext = Extension('overlay_elm_package', sources=['src/elm_doc/native/overlay_elm_package.c'], libraries=["dl"])

setup(
    name='elm-doc',
    version='0.1.0',
    url='https://github.com/ento/elm-doc',
    license='BSD 3-clause',
    author='Marica Odagaki',
    author_email='ento.entotto@gmail.com',
    description='Generate static documentation for your Elm project',
    long_description=__doc__,
    zip_safe=False,
    platforms='any',
    install_requires=[
        'click',
        'doit',
        'python-magic',
        'typing ; python_version < "3.5"',
    ],
    packages=find_packages('src', exclude=['tests']),
    package_dir={'': 'src'},
    package_data={'': ['elm_doc/native/*.js']},
    include_package_data=True,
    ext_modules=[ext],
    entry_points={
        'console_scripts': [
            'elm-doc = elm_doc.cli:main',
        ],
    },
    classifiers=[
        # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        #'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Documentation',
    ]
)
